package com.p3p.serialfiletransfer;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;

import com.p3p.serialfiletransfer.packets.FileStartPacket;
import com.p3p.serialfiletransfer.packets.FileStreamPacket;
import com.p3p.serialfiletransfer.packets.FileTransferAck;
import com.p3p.simpleserialprotocol.SerialPacket;
import com.p3p.simpleserialprotocol.SimpleSerialProtocol;

public class SerialFileTransfer {
	
	enum TransferState {
		IDLE,
		START,
		AWAITACK,
		TRANSFER,
		COMPLETE,
		ERROR;
	}

	public static void main(String[] args) {
		System.out.print("send " + args[2] + " over " + args[0] + " @ " + args[1] + " baud \r\n");
		
		TransferState state = TransferState.START;
		SimpleSerialProtocol comms = new SimpleSerialProtocol(args[0], Integer.parseInt(args[1]));
		comms.start();
		
		FileChannel inChannel = null;
		RandomAccessFile aFile = null;
		long current_chunk = 0;
		long total_chunks = 0;
		boolean active = true;
		ByteBuffer buffer = ByteBuffer.allocate(FileStreamPacket.chunksize);
		while(active){
			try {
				
				SerialPacket packet = comms.poll();
				if(packet!=null){
					switch(packet.type()){
					case 2:
						FileTransferAck ack = new FileTransferAck(packet.data);
						switch(ack.ack_type){
						case 0:
							System.out.printf("Transfer File Packet Acked %s B/s (%d of %d)\r", comms.getBpsOut(), current_chunk, total_chunks);
							state = TransferState.TRANSFER;
							break;
						case 1:
							state = TransferState.COMPLETE;
							System.out.print("Transfer File Packet Stream Acked\r\n");
							break;
						case 2:
							System.out.print("Last Packet Resend Requested\r\n");
							if(inChannel != null){
								if(inChannel.position() >= FileStreamPacket.chunksize){
									inChannel.position(inChannel.position() - FileStreamPacket.chunksize);
									state = TransferState.TRANSFER;
									break;
								}
							}
							state = TransferState.ERROR;
							break;
						case 3:
							System.out.print("Transfer Failed\r\n");
							state = TransferState.ERROR;
							break;
						default:
							state = TransferState.ERROR;
						}
						
						break;
					default:
						System.out.print("Unknown Packet Received\r\n");
						
						break;
					}
				}
				
				switch(state){
				case START:
					System.out.print("Transfer File Packet Start Sent\r\n");
					try {
						aFile = new RandomAccessFile(args[2], "r");
						inChannel = aFile.getChannel();
						long filesize = inChannel.size();
						long chunks = (long)Math.ceil(((double)filesize/FileStreamPacket.chunksize));
						total_chunks = chunks;
						
						String[] parts = args[2].split("\\.");
						comms.send(new SerialPacket(new FileStartPacket(parts[0], parts[1], chunks).encodePacket()));
						
						state = TransferState.AWAITACK;
						System.out.print("Transfer File Packet Start Awaiting Ack\r\n");
					} catch (FileNotFoundException e){
						System.out.print("File Not Found\r\n");
						state = TransferState.IDLE;
					}
					break;
				case TRANSFER:
					int bytes_read = inChannel.read(buffer);
			        if(bytes_read > 0) {
			        	long chunkid = (long)Math.ceil((double)inChannel.position() / FileStreamPacket.chunksize);
			        	current_chunk = chunkid;
			            buffer.flip();
			            comms.send(new SerialPacket(new FileStreamPacket((int)chunkid, buffer.array(), bytes_read).encodePacket()));
			            buffer.clear(); // do something with the data and clear/compact it.
			        }
			        state = TransferState.AWAITACK;
					break;
				case COMPLETE:
					inChannel.close();
			        aFile.close();
			        System.out.print("Transfer File Complete\r\n");
			        state = TransferState.IDLE;
			        active = false;
					break;
				case ERROR:
					System.out.print("Transfer Error\r\n");
					Thread.currentThread().interrupt();
					break;
				default:
					break;
				}
				
				Thread.sleep(1);

			} catch (IOException e) {
				e.printStackTrace();
			} catch (InterruptedException e){
				Thread.currentThread().interrupt();
				active = false;
				break;
			}
		}
		
		comms.interrupt();
		System.out.print("File Send Complete, Shuting Down");
	}
	
}
